import configparser
import pathlib
import time
import typing
from abc import ABC, abstractmethod
from os import environ
from threading import Thread

import requests
from pydantic import BaseModel, SecretStr


class AWSCredentials(BaseModel):
    aws_access_key: SecretStr
    aws_secret_key: SecretStr
    aws_session_token: SecretStr | None = None


class CredentialManagerBase(ABC):
    @abstractmethod
    def get_credentials(self) -> AWSCredentials:
        return AWSCredentials(aws_access_key="foo", aws_secret_key="bazz")


class CredentialManagerEnv(CredentialManagerBase):
    def __init__(
        self,
        *,
        access_key_var_name: str,
        secret_key_var_name: str,
        session_token_var_name: str,
        refresh_after: int = None,
    ):
        self.__access_key_var_name = access_key_var_name
        self.__secret_key_var_name = secret_key_var_name
        self.__session_token_var_name = session_token_var_name
        self.__refresh_after = refresh_after

        self._set_credentials()
        if refresh_after:
            thread = Thread(target=self._refresh_credentials)
            thread.daemon = True
            thread.start()

    def _refresh_credentials(self):
        while True:
            time.sleep(self.__refresh_after)
            self._set_credentials()

    def _set_credentials(self):
        self.__credentials = AWSCredentials(
            aws_access_key=environ[self.__access_key_var_name],
            aws_secret_key=environ[self.__secret_key_var_name],
            aws_session_token=environ[self.__session_token_var_name],
        )

    def get_credentials(self) -> AWSCredentials:
        return self.__credentials


class CredentialManagerArgs(CredentialManagerBase):
    def __init__(
        self, *, aws_access_key: str, aws_secret_key: str, aws_session_token: str = None
    ):
        self.__credentials = AWSCredentials(
            aws_access_key=aws_access_key,
            aws_secret_key=aws_secret_key,
            aws_session_token=aws_session_token,
        )

    def get_credentials(self) -> AWSCredentials:
        return self.__credentials


class CredentialManagerFile(CredentialManagerBase):
    def __init__(
        self,
        *,
        file_path: typing.Union[str, pathlib.Path] = "~/.aws/credentials",
        profile_name: str = "default",
        refresh_after: int | None = None,
    ):
        if isinstance(file_path, str):
            file_path = pathlib.Path(file_path)

        self.__file_path = file_path.expanduser().resolve()
        self.__profile_name = profile_name
        self.__refresh_after = refresh_after

        self._set_credentials()
        if refresh_after:
            thread = Thread(target=self._refresh_credentials)
            thread.daemon = True
            thread.start()

    def _refresh_credentials(self):
        while True:
            time.sleep(self.__refresh_after)
            self._set_credentials()

    def _set_credentials(self):
        config_parser = configparser.ConfigParser()
        with open(self.__file_path, "r") as f:
            config_parser.read_file(f)

        self.__credentials = AWSCredentials(
            aws_access_key=config_parser.get(self.__profile_name, "aws_access_key_id"),
            aws_secret_key=config_parser.get(
                self.__profile_name, "aws_secret_access_key"
            ),
            aws_session_token=config_parser.get(
                self.__profile_name, "aws_session_token", fallback=None
            ),
        )

    def get_credentials(self) -> AWSCredentials:
        return self.__credentials


class CredentialManagerMetadataService(CredentialManagerBase):
    def __init__(
        self, url: str = "http://169.254.169.254", refresh_after: int | None = None
    ):
        self._url = url

        self.__refresh_after = refresh_after

        self._set_credentials()

        if refresh_after:
            thread = Thread(target=self._refresh_credentials)
            thread.daemon = True
            thread.start()

    def _refresh_credentials(self):
        while True:
            time.sleep(self.__refresh_after)
            self._set_credentials()

    def _set_credentials(self):
        token_resp = requests.put(
            f"{self._url}/latest/api/token)",
            headers={"X-aws-ec2-metadata-token-ttl-seconds": "10"},
        )
        if token_resp.status_code == 200:
            headers = {"X-aws-ec2-metadata-token": token_resp.text}
        else:
            headers = {}
        resp = requests.get(
            f"{self._url}/latest/meta-data/iam/security-credentials/", headers=headers
        )
        role = resp.text
        resp = requests.get(
            f"{self._url}/latest/meta-data/iam/security-credentials/{role}",
            headers=headers,
        )
        resp_dict = resp.json()
        self.__credentials = AWSCredentials(
            aws_access_key=resp_dict["AccessKeyId"],
            aws_secret_key=resp_dict["SecretAccessKey"],
            aws_session_token=resp_dict.get("Token", None),
        )

    def get_credentials(self) -> AWSCredentials:
        return self.__credentials
