from __future__ import annotations

from typing import Any, Optional, Type, Union

import orjson
import pydantic

from .constants import STRING, str_to_type_mapping
from .initialization import get_table_data

try:
    import cysimdjson

    _parser = cysimdjson.JSONParser()

    def parse(raw_input: bytes) -> dict:
        resp = _parser.parse(raw_input)
        return resp.export()

    def loads(raw_input: str) -> dict:
        resp = _parser.loads(raw_input)
        return resp.export()

except ModuleNotFoundError:

    def parse(raw_input: bytes) -> dict:
        return orjson.loads(raw_input)

    def loads(raw_input: str) -> dict:
        return orjson.loads(raw_input)


_delimiter = ";"


def orjson_dumps(v, *, default):
    return orjson.dumps(v, default=default).decode()


def fix_item(cls, input_: dict):
    for k, v in input_.items():
        if isinstance(v, dict):
            for k1, v1 in v.items():
                if k1 in str_to_type_mapping:
                    input_[k] = v1
        else:
            input_[k] = v
    return input_


def _add_dict(d):
    new_dict = {}
    for k, v in d.items():
        if not isinstance(v, pydantic.BaseModel):
            new_dict[k] = {STRING: v}
        else:
            new_dict[k] = {STRING: v.json()}
    return new_dict


class TableBaseModel(pydantic.BaseModel):
    class DoesNotExist(Exception):
        pass

    def __init__(self, **data: Any):
        self._setup_class()
        super().__init__(**data)

    def __get__(self, instance, owner):
        """
        needed only to ease coding.
        temp solution for pycharm bug : https://youtrack.jetbrains.com/issue/PY-36176
        """
        super(__class__, self).__get__(instance, owner)
        return self

    def save(self):
        table_data = get_table_data()
        hash_key_value = self.Config.hash_key_attribute

        payload = {
            table_data.hash_key: {
                STRING: f"{self.__class__.__name__}{_delimiter}{self.__getattribute__(hash_key_value)}"
            },
        }

        if table_data.range_key and self.Config.range_key_attribute:
            range_key_value = self.Config.range_key_attribute
            payload[table_data.range_key] = {
                STRING: f"{self.__class__.__name__}{_delimiter}{self.__getattribute__(range_key_value)}"
            }

        payload.update(self._dydb_dump())
        table_data.dydb.put_item(TableName=table_data.table_name, Item=payload)

    def _dydb_dump(self) -> dict:
        json_dict = self.dict()
        return _add_dict(json_dict)

    @classmethod
    def get(cls, *, hash_key, range_key: Optional[str]):
        cls._setup_class()
        table_data = get_table_data()

        key = {
            table_data.hash_key: {STRING: f"{cls.__name__}{_delimiter}{hash_key}"},
        }

        if table_data.range_key and cls.Config.range_key_attribute:
            key[table_data.range_key] = {
                STRING: f"{cls.__name__}{_delimiter}{range_key}"
            }

        resp = table_data.dydb.get_item(
            TableName=table_data.table_name,
            Key=key,
            AttributesToGet=cls.Config.attr_to_get,
        )

        if resp == b"{}":
            raise cls.DoesNotExist(
                f"Item with hash key:{hash_key} and range key:{range_key} was not found"
            )

        data = parse(resp)["Item"]
        return cls.parse_obj(data)

    @classmethod
    def batch_get(cls, items: Union[list[str] | list[tuple[str, str]]]):
        # todo implement multiple batch calls if partial results obtained
        table_data = get_table_data()
        requests = []

        for item in items:
            if isinstance(item, tuple):
                primary_value = f"{cls.__name__}{_delimiter}{item[0]}"
            else:
                primary_value = f"{cls.__name__}{_delimiter}{item}"

            if table_data.range_key and cls.Config.range_key_attribute:
                secondary_value = f"{cls.__name__}{_delimiter}{item[1]}"
                request = {
                    table_data.hash_key: {STRING: primary_value},
                    table_data.range_key: {STRING: secondary_value},
                }

            else:
                request = {table_data.hash_key: {STRING: primary_value}}
            requests.append(request)

        chunks = _chunks(requests, 100)

        for chunk in chunks:
            resp = table_data.dydb.batch_get(
                RequestItems={
                    table_data.table_name: {
                        "AttributesToGet": cls.Config.attr_to_get,
                        "Keys": chunk,
                    }
                }
            )
            items = parse(resp)["Responses"][table_data.table_name]
            for item in items:
                data = cls.parse_obj(item)
                yield data

    @classmethod
    def scan(cls):
        # TODO implement scan
        # todo implement multiple scan calls if not all values obtained
        # todo implement retries if request limit exceeded
        cls._setup_class()
        table_data = get_table_data()
        paginator = table_data.dydb_client.get_paginator("scan")
        resp_iterator = paginator.paginate(
            TableName=table_data.table_name,
            AttributesToGet=["_d"],
            ScanFilter={
                table_data.hash_key: {
                    "ComparisonOperator": "BEGINS_WITH",
                    "AttributeValueList": [{STRING: cls.__name__}],
                }
            },
        )
        for resp in resp_iterator:
            for item in resp["Items"]:
                yield cls.parse_raw(item["_d"][STRING])

    @classmethod
    def query(cls, primary_key_value: str, filter_range=False):
        # todo implement multiple query calls if not all values obtained
        # todo implement retries if request limit exceeded
        cls._setup_class()
        table_data = get_table_data()

        key_conditions = {
            table_data.hash_key: {
                "AttributeValueList": [
                    {STRING: f"{cls.__name__}{_delimiter}{primary_key_value}"}
                ],
                "ComparisonOperator": "EQ",
            }
        }

        if table_data.range_key and cls.Config.range_key_attribute and filter_range:
            key_conditions[table_data.range_key] = {
                "AttributeValueList": [{STRING: f"{cls.__name__}{_delimiter}"}],
                "ComparisonOperator": "CONTAINS",
            }

        resp = table_data.dydb.query(
            TableName=table_data.table_name,
            ReturnConsumedCapacity="TOTAL",
            AttributesToGet=cls.Config.attr_to_get,
            KeyConditions=key_conditions,
        )

        def get_obj(json_d):
            return cls.parse_obj(json_d)

        json_dict = parse(resp)["Items"]
        return list(map(get_obj, json_dict))

    @classmethod
    def delete(cls, primary_key_value: str, secondary_key_value: Optional[str] = None):
        cls._setup_class()
        table_data = get_table_data()

        payload = {
            table_data.hash_key: {
                STRING: f"{cls.__name__}{_delimiter}{primary_key_value}"
            },
        }

        if table_data.range_key and cls.Config.range_key_attribute:
            payload[table_data.range_key] = {
                STRING: f"{cls.__name__}{_delimiter}{secondary_key_value}"
            }

        resp = table_data.dydb.delete(
            TableName=table_data.table_name,
            Key=payload,
        )

    def delete_obj(self):
        table_data = get_table_data()
        hash_key_value = self.Config.hash_key_attribute

        payload = {
            table_data.hash_key: {
                STRING: f"{self.__class__.__name__}{_delimiter}{self.__getattribute__(hash_key_value)}"
            },
        }

        if table_data.range_key and self.Config.range_key_attribute:
            range_key_value = self.Config.range_key_attribute
            payload[table_data.range_key] = {
                STRING: f"{self.__class__.__name__}{_delimiter}{self.__getattribute__(range_key_value)}"
            }

        resp = table_data.dydb.delete(
            TableName=table_data.table_name,
            Key=payload,
        )

    @classmethod
    def _setup_class(cls):
        if not getattr(cls.Config, "hash_key_attribute", None):
            try:
                hash_key = cls.Config.hash_key
            except AttributeError:
                raise Exception(f"Required hash_key attribute not present in Config")

            try:
                hash_key_attribute = cls.__dict__["__fields__"][hash_key]
            except KeyError:
                if not isinstance(type(hash_key), str):
                    Exception(
                        f"Config class attribute hash_key is expected to be {str} but is {type(hash_key)}"
                    )
                raise Exception(
                    f"Config class attribute hash_key value: {hash_key} is not an attribute of {cls.__name__}"
                )

            if hash_key_attribute.type_ is not str:
                raise Exception(
                    f"Hash key attribute type is required to be {str} but is {hash_key_attribute.type_}"
                )

            cls.Config.hash_key_attribute = hash_key_attribute

        if not getattr(cls.Config, "range_key_attribute", None):
            range_key = getattr(cls.Config, "range_key", None)

            if range_key:
                try:
                    range_key_attribute = cls.__dict__["__fields__"][range_key]
                except KeyError:
                    if not isinstance(type(range_key), str):
                        raise Exception(
                            f"Config class attribute range_key is expected to be {str} but is {type(range_key)}"
                        )
                    raise Exception(
                        f"Config class attribute range_key: {range_key} is not an attribute of {cls.__name__}"
                    )
                if range_key_attribute.type_ is not str:
                    raise Exception(
                        f"range key attribute type is required to be {str} but is {range_key_attribute.type_}"
                    )
            else:
                range_key_attribute = None
            cls.Config.range_key_attribute = range_key_attribute

        if not getattr(cls.Config, "json_loads", None):
            cls.Config.json_loads = orjson.loads
        if not getattr(cls.Config, "json_dumps", None):
            cls.Config.json_dumps = orjson_dumps
        if not getattr(cls.Config, "attr_to_get", None):
            cls.Config.attr_to_get = [key for key in cls.__fields__.keys()]
        if not getattr(cls.Config, "_validator_set", None):
            cls.__pre_root_validators__.append(fix_item)
            cls.Config._validator_set = True


def _chunks(lst, n):
    """Yield successive n-sized chunks from lst."""
    for i in range(0, len(lst), n):
        yield lst[i : i + n]


def batch_delete(
    items: list[tuple[Type[TableBaseModel], str]]
    | list[tuple[Type[TableBaseModel], str, str]]
):
    # todo implement multiple delete calls if unprocessed items
    # todo implement retries if request limit exceeded
    table_data = get_table_data()

    requests = []
    for item in items:
        item[0]._setup_class()
        if len(item) == 3:
            request = {
                "DeleteRequest": {
                    table_data.hash_key: {
                        STRING: f"{item[0].__class__.__name__}{_delimiter}{item[1]}"
                    },
                    table_data.range_key: {
                        STRING: f"{item[0].__class__.__name__}{_delimiter}{item[2]}"
                    },
                }
            }
        else:
            request = {"DeleteRequest": {table_data.hash_key: {STRING: item[1]}}}
        requests.append(request)

    chunks = _chunks(requests, 25)

    for chunk in chunks:
        resp = table_data.dydb.batch_write(RequestItems=chunk)
    return


def batch_delete_obj(items: list[TableBaseModel]):
    # todo implement multiple delete calls if unprocessed items
    # todo implement retries if request limit exceeded
    table_data = get_table_data()

    requests = []
    for item in items:
        hash_key_value = item.Config.hash_key_attribute
        primary_value = f"{item.__class__.__name__}{_delimiter}{item.__getattribute__(hash_key_value)}"

        if table_data.range_key and item.Config.range_key_attribute:
            range_key_value = item.Config.range_key_attribute
            secondary_value = f"{item.__class__.__name__}{_delimiter}{item.__getattribute__(range_key_value)}"
            request = {
                "DeleteRequest": {
                    "Key": {
                        table_data.hash_key: {STRING: primary_value},
                        table_data.range_key: {STRING: secondary_value},
                    }
                }
            }
        else:
            request = {
                "DeleteRequest": {"Key": {table_data.hash_key: {STRING: primary_value}}}
            }
        requests.append(request)

    chunks = _chunks(requests, 25)

    for chunk in chunks:
        resp = table_data.dydb.batch_write(RequestItems={table_data.table_name: chunk})
    return


def batch_write(items: list[TableBaseModel]):
    # todo implement multiple write calls if unprocessed items
    # todo implement retries if request limit exceeded
    put_request = "PutRequest"
    table_data = get_table_data()

    requests = []

    for item in items:
        hash_key_value = item.Config.hash_key_attribute
        primary_value = f"{item.__class__.__name__}{_delimiter}{item.__getattribute__(hash_key_value)}"

        if table_data.range_key and item.Config.range_key_attribute:
            range_key_value = item.Config.range_key_attribute
            secondary_value = f"{item.__class__.__name__}{_delimiter}{item.__getattribute__(range_key_value)}"
            request = {
                put_request: {
                    "Item": {
                        table_data.hash_key: {STRING: primary_value},
                        table_data.range_key: {STRING: secondary_value},
                    }
                }
            }
        else:
            request = {
                put_request: {"Item": {table_data.hash_key: {STRING: primary_value}}}
            }

        request[put_request].update(item._dydb_dump())

        requests.append(request)

    chunks = _chunks(requests, 25)

    for chunk in chunks:
        table_data.dydb.batch_write(RequestItems={table_data.table_name: chunk})
    return
